@@ -1,5 +1,5 @@
 /* PSPP - a program for statistical analysis.
-   Copyright (C) 2008 Free Software Foundation, Inc.
+   Copyright (C) 2008, 2011 Free Software Foundation, Inc.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@
 #define PSQL_READER_H 1
 
 #include <stdbool.h>
-#include <libpspp/str.h>
+#include "libpspp/str.h"
 
 struct casereader;
 
