@@ -117,18 +117,6 @@ psppire_buttonbox_get_property (GObject         *object,
 }
 
 
-typedef enum
-  {
-    PSPPIRE_BUTTON_OK_MASK     = (1 << PSPPIRE_BUTTON_OK),
-    PSPPIRE_BUTTON_GOTO_MASK   = (1 << PSPPIRE_BUTTON_GOTO),
-    PSPPIRE_BUTTON_CONTINUE_MASK = (1 << PSPPIRE_BUTTON_CONTINUE),
-    PSPPIRE_BUTTON_CANCEL_MASK = (1 << PSPPIRE_BUTTON_CANCEL),
-    PSPPIRE_BUTTON_CLOSE_MASK  = (1 << PSPPIRE_BUTTON_CLOSE),
-    PSPPIRE_BUTTON_HELP_MASK   = (1 << PSPPIRE_BUTTON_HELP),
-    PSPPIRE_BUTTON_RESET_MASK  = (1 << PSPPIRE_BUTTON_RESET),
-    PSPPIRE_BUTTON_PASTE_MASK  = (1 << PSPPIRE_BUTTON_PASTE)
-  } PsppireButtonMask;
-
 static GParamSpec *button_flags;
 static GParamSpec *default_flags;
 
@@ -416,3 +404,8 @@ psppire_button_flags_get_type (void)
   return ftype;
 }
 
+GtkWidget*
+psppire_button_box_new (void)
+{
+  return GTK_WIDGET (g_object_new (psppire_button_box_get_type (), NULL));
+}
